/*
  Daemonshogi -- a GTK+ based, simple shogi (Japanese chess) program.

  Copyright (C) Masahiko Tokita    2002-2005,2009 
  Copyright (C) Hisashi Horikawa   2008-2010

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "board-misc.h"
#include <string>
#include <string.h>
using namespace std;


/** (x, y)に駒 p をセットする。 */
void daemon_dboard_set_board(DBoard* bo, int x, int y, PieceKind p) 
{
  assert(bo != NULL);
  assert(1 <= x && x <= 9);
  assert(1 <= y && y <= 9);
  assert(0 <= p && p <= 0x1F);

  bo->board[(y << 4) + x] = p;
}


/** (x, y)の駒を返す。 */
PieceKind daemon_dboard_get_board(const DBoard* bo, int x, int y) 
{
  assert(bo != NULL);
  assert(1 <= x && x <= 9);
  assert(1 <= y && y <= 9);

  return bo->board[(y << 4) + x];
}


/** next 手番の持ち駒 p の枚数をセットする。 */
void daemon_dboard_set_piece(DBoard* bo, int next, PieceKind p, int n) 
{
  bo->inhand[next][p & 7] = n;
}


/** next 手番の持ち駒 p の枚数を一枚プラスする。 */
void daemon_dboard_add_piece(DBoard* bo, int next, PieceKind p) 
{
  bo->inhand[next][p & 7]++;
}


/** next 手番の持ち駒 p の枚数を一枚減らす。 */
void daemon_dboard_dec_piece(DBoard* bo, int next, PieceKind p) 
{
  bo->inhand[next][p & 7]--;
  assert( bo->inhand[next][p & 7] >= 0 );
}


/**
 * 持ち駒 p の枚数を返す。
 * @param bo 局面
 * @param next 手番
 * @param p 駒種
 */
int daemon_dboard_get_piece(const DBoard* bo, int next, PieceKind p) 
{
  return bo->inhand[next][p & 7];
}


/** 平手の盤面をセットする */
void bo_reset_to_hirate( BOARD* bo )
{
  // 平手の配置パターン
  static const int hirate[] = {
    0x12, 0x13, 0x14, 0x15, 0x18, 0x15, 0x14, 0x13, 0x12,
    0x00, 0x16, 0x00, 0x00, 0x00, 0x00, 0x00, 0x17, 0x00,
    0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
    0x00, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00,
    0x02, 0x03, 0x04, 0x05, 0x08, 0x05, 0x04, 0x03, 0x02,
  };
  int x, y;
  
  bo_init(bo);
  
  for (y = 0; y < 9; y++) {
    for (x = 0; x < 9; x++) {
      daemon_dboard_set_board(bo, x + 1, y + 1, hirate[y * 9 + x]);
    }
  }

  // 利きを更新
  boSetToBOARD( bo );
}


/** 詰将棋用に盤面をセットする */
void daemon_dboard_set_mate(DBoard* board) 
{
  bo_init(board);

  daemon_dboard_set_board(board, 5, 1, 0x18);
  boSetTsumeShogiPiece( board );
}


void daemon_dboard_set_rl_reversal(DBoard *board) 
{
  int x, y, p;

  BOARD* tmp = bo_dup(board);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(tmp, x, y);
      daemon_dboard_set_board(board, 10 - x, y, p);
    }
  }

  freeBOARD( tmp );
}


void daemon_dboard_set_order_reversal(DBoard* board) 
{
  int x, y;
  PieceKind p;

  BOARD* tmp = bo_dup(board);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(tmp, x, y);
      if (p != 0) {
	p ^= 0x10;
      }
      daemon_dboard_set_board(board, x, 10 - y, p);
    }
  }

  freeBOARD( tmp );
}


/** 先手後手の盤面を入れ換える */
void daemon_dboard_flip(DBoard* board) 
{
  int x, y, i;
  PieceKind p;

  BOARD* tmp = bo_dup(board);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(tmp, x, y);
      if (p != 0) {
	p ^= 0x10;
      }
      daemon_dboard_set_board(board, 10 - x, 10 - y, p);
    }
  }

  for (i=1; i<=7; i++) {
    p = daemon_dboard_get_piece(tmp, SENTE, i);
    daemon_dboard_set_piece(board, GOTE, i, p);
    p = daemon_dboard_get_piece(tmp, GOTE, i);
    daemon_dboard_set_piece(board, SENTE, i, p);
  }

  freeBOARD( tmp );
}


static const int piece_max[] = {
  0, 18, 4, 4, 4, 4, 2, 2, 2,
};

/** 全ての駒を駒箱へ移動する */
void daemon_dboard_set_all_to_pbox(DBoard* board) 
{
  int x, y, i;

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      daemon_dboard_set_board(board, x, y, 0);
    }
  }

  for (i=1; i<=7; i++) {
    daemon_dboard_set_piece(board, GOTE, i, 0);
    daemon_dboard_set_piece(board, SENTE, i, 0);
  }

  for (i=1; i<=8; i++) {
    daemon_dboard_set_pbox(board, i, piece_max[i]);
  }
}


/** 編集用に使っていない駒を pbox に格納する */
void daemon_dboard_set_for_edit(DBoard *board) 
{
  int piece[9];
  int x, y, p;
  int i;
  
  for (i=0; i<9; i++) {
    piece[i] = 0;
  }

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(board, x, y);
      if ((p & 0xF) == 8) {
	p = 8;
      } else {
	p &= 7;
      }
      piece[p]++;
    }
  }

  for (i=1; i<=7; i++) {
    p = daemon_dboard_get_piece(board, SENTE, i);
    piece[i] += p;
    p = daemon_dboard_get_piece(board, GOTE, i);
    piece[i] += p;
  }

  for (i=1; i<=8; i++) {
    daemon_dboard_set_pbox(board, i, piece_max[i] - piece[i]);
  }
}


/** DBoard の中見を src から dest にコピーする */
void daemon_dboard_copy(const DBoard* src, DBoard *dest) 
{
  *dest = *src;
}


void daemon_dboard_set_pbox(DBoard *bo, PieceKind p, int n) 
{
  assert(bo != NULL);
  assert(1 <= p && p <= 8);
  assert(0 <= n && n <= 18);

  bo->pbox[p] = n;
}


int daemon_dboard_get_pbox(const DBoard* bo, PieceKind p) 
{
  assert(bo != NULL);
  assert(1 <= p && p <= 8);

  return bo->pbox[p];
}

void daemon_dboard_add_pbox(DBoard *bo, int p) {
  assert(bo != NULL);
  assert(1 <= p && p <= 8);

  bo->pbox[p]++;
}

void daemon_dboard_dec_pbox(DBoard *bo, int p) {
  assert(bo != NULL);
  assert(1 <= p && p <= 8);

  bo->pbox[p]--;
}


/**
 * 読み込んだ1行から局面を更新する。
 * @param bo 局面
 * @param s 1行
 * @param error エラーがあったときにエラーを知らせる。NULLでもよい。
 */
void bo_update_board_by_csa_line(BOARD* bo, const char* s, int* error)
{
  assert(s);

  int side;
  int i, y;

  if (*s == '+') {
    bo->next = SENTE;
    return;
  }
  else if (*s == '-') {
    bo->next = GOTE;
    return;
  }

  s++;
  if (*s == '+' || *s == '-') {
      /* 持ち駒 */
      side = *s == '+' ? 0 : 1;
      s++;

      while (*s != '\0') {
	if (*s != '0' || *(s + 1) != '0') {
	  /* 仕様上ありえない */
          if (error)
            *error = 1; // record->set_error("mochigoma separator error");
	  return;
	}
	s += 2;
	for (i = 1; i <= 7; i++) {
          if (strncmp(s, CSA_PIECE_NAME[i], 2) == 0)
	    break;
	}
	if (i > 7) {
	  /* 仕様上ありえない */
          if (error)
            *error = 2; // record->set_error("mochigoma too much");
	  return;
	}
        bo->inhand[side][i]++;
	s += 2;
      }
    }
    else if (*s >= '1' && *s <= '9') {
      // 盤面
#if 0
      if (y != *s - 0x30) {
	/* 仕様上ありえない */
        if (record)
          record->set_error("board lineno error");
	return;
      }
#endif // 0
      y = *s - 0x30;
      s++;

      for (int x = 9; x >= 1; x--) {
	if (*s == ' ') {
          bo->board[(y << 4) + x] = 0;
	  s += 3;
	  continue;
	}
	else if (*s == '+')
	  side = 0; // 先手
	else if (*s == '-')
	  side = 1; // 後手
	else {
          if (error)
            *error = 3; // record->set_error("piece side error");
	  return;
	}
	s++;

	for (i = 1; i <= 15; i++) {
          if (strncmp(s, CSA_PIECE_NAME[i], 2) == 0)
	    break;
	}
	if (i > 15) {
	  /* 仕様上ありえない */
          if (error)
            *error = 4; // record->set_error("piece kind error");
	  return;
	}
        bo->board[(y << 4) + x] = i + side * 0x10;
	s += 2;
      }
      // y++;
    }
    else {
      // 持ち駒でも盤面でもない
      if (error)
        *error = 5; // record->set_error("unknown line type");
      return;
    }
}


/**
 * 文字列から局面をセットする
 * @return 成功はtrue
 */
bool bo_set_by_text(BOARD* bo, const char* s)
{
  string text = s;
  int idx = 0;
  string::size_type epos;
  int error = 0;

  bo_init(bo);

  // 持ち駒を初期化
  for (int i = 0; i <= 7; i++) {
    bo->inhand[0][i] = 0;
    bo->inhand[1][i] = 0;
  }

  while ( (epos = text.find("\n", idx)) != string::npos ) {
    string line(text, idx, epos - idx);
    if (line[0] == '\'') {
      idx = epos + 1;
      continue;
    }

    bo_update_board_by_csa_line(bo, line.c_str(), &error);
    if (error)
      return false;

    idx = epos + 1;
  }

  // 枚数が多すぎないか
  if (!boCheckJustMate(bo)) {
    printf("%s: board error!!!!!\n", __func__); // DEBUG
    return false;
  }

  // 利き情報を更新
  boSetToBOARD(bo);

  return true;
}


/** 
 * CSAファイルから局面を読み込む
 * @return 成功は true
 */
bool bo_load_file( BOARD* bo, const char* filename )
{
  FILE* fp = fopen(filename, "r");
  if (!fp)
    return false;

  char buf[2000];
  int r = fread(buf, 1, sizeof(buf) - 1, fp);
  buf[r] = '\0';

  fclose(fp);

  if ( !bo_set_by_text(bo, buf) )
    return false;

  // 利きを更新
  boSetToBOARD(bo);

  return true;
}
